/*
 * This file was generated by the mksignames program.
 * (And then later tailored for MiNT ST by Stephan Neuhaus
 * neuhaus@informatik.uni-kl.de)
 */

#include "shell.h"

char *const sigmesg[32] = {
      0,				/*  0 */
      "Hangup",				/*  1 */
      "Interrupt",			/*  2 */
      "Quit",				/*  3 */
      "Illegal instruction",		/*  4 */
      "Trace/BPT trap",			/*  5 */
      "abort",				/*  6 */
      "Privilege violation",		/*  7 */
      "Floating exception",		/*  8 */
      "Killed",				/*  9 */
      "Bus error",			/* 10 */
      "Segmentation fault",		/* 11 */
      "Bad system call",		/* 12 */
      "Broken pipe",			/* 13 */
      "Alarm call",			/* 14 */
      "Terminated",			/* 15 */
      "Urgent I/O",			/* 16 */
      "Stopped",			/* 17 */
      "Stopped",			/* 18 */
      "Continue",			/* 19 */
      "Child stopped or exited",	/* 20 */
      "Stopped (input)",		/* 21 */
      "Stopped (output)",		/* 22 */
      "I/O possible",			/* 23 */
      "Time limit exceeded",		/* 24 */
      "File size limit exceeded",	/* 25 */
      "Virtual timer alarm",		/* 26 */
      "Profiling alarm",		/* 27 */
      "Window size changed",		/* 28 */
      "User signal 1",			/* 29 */
      "User signal 2",			/* 30 */
      "Unknown signal",			/* 31 */
};
